
#include <QSettings>
#include <QVBoxLayout>
#include <QModelIndex>
#include <QHeaderView>
#include <QMessageBox>
#include <QCloseEvent>
#include <QApplication>
#include <iostream>

#ifdef WIN32
  #include <C:\Boost\include\boost-1_33_1\boost\lexical_cast.hpp>
#else
  #include "boost/lexical_cast.hpp"
#endif


#include "AddDatabase.h"
#include "Const.h"

using boost::lexical_cast;
using boost::timer;



// Thread instance for elapsed time of Adding a database
MyThread::MyThread(SetupPage* setupPage)
{
  setupPage_ = setupPage;
}


// Starts the Timer in a thread and signals to update the display with time elapsed once a second
void MyThread::run()
{
  QTimer *timer = new QTimer(this);
  connect(timer, SIGNAL(timeout()), this, SLOT(updateTime()));
  timer->start(1000);
  exec();
}


void MyThread::updateTime()
{
  setupPage_->updateTime();
}



AddDatabase::AddDatabase(MainWindow* mainWindow, QWidget *parent) : 
    QDialog(parent), mainWindow_(mainWindow)
{
  setupPage = new SetupPage(this);
  progressPage = new ProgressPage(this);
  stackedWidget = new QStackedWidget;
  stackedWidget->addWidget(setupPage);
  stackedWidget->addWidget(progressPage);
  stackedWidget->setCurrentWidget(setupPage);  
  QVBoxLayout* layout = new QVBoxLayout;
  layout->addWidget(stackedWidget);
  setLayout(layout);
  setWindowTitle(tr("Add Database"));
  setWindowIcon(QIcon(":/resources/icons/misc.png"));
  setWindowModality(Qt::ApplicationModal);
  readSettings();
}



bool AddDatabase::isCancelled()
{
  return progressPage->hasCancelled_;
}



void AddDatabase::updateEvents()
{
  qApp->processEvents();
}



void AddDatabase::reject()
{
  writeSettings();
  setResult(QDialog::Rejected);
  hide();
}



void AddDatabase::accept()
{
  writeSettings();
  setResult(QDialog::Accepted);
  hide();
}
    

    
void AddDatabase::readSettings()
{
  QSettings settings;
  settings.beginGroup("AddDatabase");
  QPoint pos = settings.value("pos", QPoint(100, 100)).toPoint();
  QSize size = settings.value("size", QSize(300, 600)).toSize();                   
  settings.endGroup();  
  resize(size);
  move(pos);
}



void AddDatabase::writeSettings()
{
  QSettings settings;
  settings.beginGroup("AddDatabase");
  settings.setValue("pos", pos());
  settings.setValue("size", size());
  settings.endGroup();
}



//--------------------------------------------------------------------------
// SETUP PAGE
//--------------------------------------------------------------------------
SetupPage::SetupPage(AddDatabase* addDatabase, QWidget* parent) : QWidget(parent)
{
  addDatabase_ = addDatabase;
  databaseNameLabel = new QLabel(tr("Choose Database &name:"));
  databaseNameLineEdit = new QLineEdit;
  databaseNameLabel->setBuddy(databaseNameLineEdit);
  setFocusProxy(databaseNameLineEdit);

  directoryNameLabel = new QLabel(tr("Select a Directory with Sgf's :"));
  dirModel = new QDirModel;
  dirModel->setFilter(QDir::Dirs);
  dirModel->setReadOnly(true);
  dirModel->setLazyChildCount(false);
  treeView = new QTreeView;
  treeView->setModel(dirModel);
  treeView->header()->resizeSection(0,200);
  treeView->setColumnHidden(0, false); // Show only dir name
  treeView->setColumnHidden(1, true); // Show only dir name
  treeView->setColumnHidden(2, true); // Not type, size, modified
  treeView->setColumnHidden(3, true);
  
  treeView->setCurrentIndex (dirModel->index(QDir::currentPath()));
  groupBox = new QGroupBox(tr("&Options"));
  recursiveDirCheckBox = new QCheckBox(tr("&Search directories recursively."));
  acceptDuplicatesCheckBox = new QCheckBox(tr("&Reject duplicates."));
  acceptDuplicatesCheckBox->setCheckState (Qt::Checked);

  QVBoxLayout *groupBoxLayout = new QVBoxLayout;
  groupBoxLayout->addWidget(recursiveDirCheckBox);
  groupBoxLayout->addWidget(acceptDuplicatesCheckBox);
  groupBox->setLayout(groupBoxLayout);

  cancelButton = new QPushButton("&Cancel", this);
  startButton = new QPushButton("&Start", this);

  QGridLayout *layout = new QGridLayout;
  layout->setRowMinimumHeight(0, 10);
  layout->addWidget(databaseNameLabel, 0, 0);
  layout->addWidget(databaseNameLineEdit, 0, 1);
  layout->addWidget(directoryNameLabel, 1, 0, 1, 2);
  layout->addWidget(treeView, 2, 0, 1, 2);
  layout->addWidget(groupBox, 3, 0, 1, 2);

  QSpacerItem* spacer = new QSpacerItem(50, 20);
  QGridLayout *layoutButtons = new QGridLayout;
  layoutButtons->addItem(spacer, 0, 0);
  layoutButtons->addWidget(cancelButton, 0, 1);
  layoutButtons->addWidget(startButton, 0, 2);
  layoutButtons->setColumnStretch(0, 1);
  layout->addItem(layoutButtons, 4, 0, 1, 2);
  
  layout->setOriginCorner(Qt::TopLeftCorner);
  setLayout(layout);
  
  connect(databaseNameLineEdit, SIGNAL(textChanged(const QString &)),
          this, SLOT(checkFirstPage()));
  connect(treeView, SIGNAL(clicked(const QModelIndex &)),
          this, SLOT(checkFirstPage()));
  connect(cancelButton, SIGNAL(clicked()),
          addDatabase_, SLOT(reject()));
  connect(startButton, SIGNAL(clicked()),
          this, SLOT(start()));
  
  startButton->setEnabled(false);  
}



void SetupPage::checkFirstPage()
{
  QString txt = dirModel->filePath(treeView->currentIndex());
  if ((databaseNameLineEdit->text().isEmpty() == false) && (txt.isEmpty() == false)) {
    startButton->setEnabled(true);
  }
  else {
    startButton->setEnabled(false);
  }
}



/*******************************************************************88
  Start the Process of Adding a new Database
  Start the timer, change cursor, 
  Find out if we can create the database using the selected name
  If we can create the database then we change to the Progress Page 
  Then start Database->setupDatabase process.
**********************************************************************/
void SetupPage::start()
{
  timeSeconds_ = 0;
  MyThread* myThread = new MyThread(this);
  myThread->start();
  
  qApp->setOverrideCursor(QCursor(Qt::WaitCursor));
  addDatabase_->timer_ = new Timer; // Start timer 
  bool recursive = false;
  if (recursiveDirCheckBox->checkState() == Qt::Checked) {
    recursive = true;
  }
  bool eraseDuplicates = false;
  if (acceptDuplicatesCheckBox->checkState() == Qt::Checked) {
    eraseDuplicates = true;
  }
  
  QString name = databaseNameLineEdit->text();
  QModelIndex index = treeView->currentIndex();
  QString path = dirModel->filePath(index);
  setupDatabaseError result = 
      addDatabase_->mainWindow_->dbManager_->addDatabase(name.toStdString(),
      path.toStdString(), recursive);

  switch (result) {
    case NAME_TAKEN : {
      qApp->restoreOverrideCursor();
      myThread->quit();
      QMessageBox::information(this, "baduK", "Unable to create database with given name.\n" "Choose a different Database name.");
      return;
    }
    case DIR_TAKEN : {
      qApp->restoreOverrideCursor();
      myThread->quit();
      QMessageBox::information(this, "baduK", "Unable to create database directory.\n"
          "Choose a different Database name.");      
      return;
    }
    case ACCEPT : {
      break;
    }
    case NO_SGF : {
      qApp->restoreOverrideCursor();
      myThread->quit();
      QMessageBox::information(this, "messageDlg", "Specified Directory has no SGF files.\n"
          "Choose a different Directory.");
      QString name = databaseNameLineEdit->text();
      std::string strName = name.toStdString();
      addDatabase_->mainWindow_->dbManager_->deleteDatabase(strName);
      return;
    }
    case BAD_DIRECTORY : {
      qApp->restoreOverrideCursor();
      myThread->quit();
      QMessageBox::information(this, "baduK", "Specified Directory does not exist.\n"
          "Choose a different Directory.");
      
      if (ACCEPT != result) {
        QString name = databaseNameLineEdit->text();
        std::string strName = name.toStdString();
        addDatabase_->mainWindow_->dbManager_->deleteDatabase(strName);
      }
      return;
    }
    case UNABLE_TO_CREATE_DIR : {}
    case BAD_SGF : {}
    case CANCEL : {}
  }
  addDatabase_->stackedWidget->setCurrentWidget(addDatabase_->progressPage);
  addDatabase_->progressPage->labelCurrentTimeData->setText("0:00 Seconds");
  addDatabase_->repaint();
  addDatabase_->update();
  
  result = addDatabase_->mainWindow_->dbManager_->setupDatabase(
      addDatabase_->mainWindow_, eraseDuplicates);
  qApp->restoreOverrideCursor();
  myThread->quit();

  switch (result) {
    case NAME_TAKEN : {}
    case DIR_TAKEN : {}
    case BAD_DIRECTORY : {}
    case NO_SGF : {}
    case UNABLE_TO_CREATE_DIR : {}
    case BAD_SGF : {}
    case ACCEPT : {
      addDatabase_->progressPage->finishButton->setEnabled(true);
      break;
    }
    case CANCEL : {
      if (true == addDatabase_->progressPage->isBackClicked_) {
        addDatabase_->progressPage->back();
      }
      else {
        addDatabase_->progressPage->cancel();
      }
      break;
    }
  }
}



// Display new current time
void SetupPage::updateTime()
{
  int minutes = 0;
  ++timeSeconds_;
  double dblSeconds = timeSeconds_;
  int remainderSeconds = timeSeconds_;
  if (timeSeconds_ >= 60) {
    minutes = static_cast<int>(dblSeconds / 60);
    remainderSeconds = timeSeconds_ - (minutes * 60);
  };
  std::string time = lexical_cast<std::string>(minutes) + ":";
  if (remainderSeconds < 10) {
    time += "0";
  }
  time += lexical_cast<std::string>(remainderSeconds) + " Seconds";
  addDatabase_->progressPage->labelCurrentTimeData->setText(time.c_str());
  addDatabase_->repaint();
  addDatabase_->update();
}



//--------------------------------------------------------------------------
// PROGRESS PAGE
//--------------------------------------------------------------------------
ProgressPage::ProgressPage(AddDatabase* addDatabase, QWidget* parent) : QWidget(parent)
{
  addDatabase_ = addDatabase;
  labelInitializeGames = new QLabel("Initialize Games");
  labelDuplicateCheck = new QLabel("Duplicate Check");
  labelParseGames = new QLabel("Parsing Games");
  labelInitializePatterns = new QLabel("Generating Patterns");
  labelOverall = new QLabel("Overall");
  labelInitializeGames->setMaximumHeight(20);
  labelDuplicateCheck->setMaximumHeight(20);
  labelParseGames->setMaximumHeight(20);
  labelInitializePatterns->setMaximumHeight(20);
  labelOverall->setMaximumHeight(20);
  
  labelCurrentTime = new QLabel("Elapsed time");
  labelEstimateTime = new QLabel("Estimated time");
  labelTotalGames = new QLabel("Number of Games found");
  labelDuplicateGames = new QLabel("Duplicate Games found");
  
  labelCurrentTime->setMaximumHeight(20);
  labelEstimateTime->setMaximumHeight(20);
  labelTotalGames->setMaximumHeight(20);
  labelDuplicateGames->setMaximumHeight(20);
  
  labelCurrentTimeData = new QLabel;
  labelEstimateTimeData = new QLabel;
  labelTotalGamesData = new QLabel;
  labelDuplicateGamesData = new QLabel;
    
  labelCurrentTimeData->setMaximumHeight(20);
  labelEstimateTimeData->setMaximumHeight(20);
  labelTotalGamesData->setMaximumHeight(20);
  labelDuplicateGamesData->setMaximumHeight(20);
  
  progressBarInitializeGames = new QProgressBar;
  progressBarDuplicateCheck = new QProgressBar;
  progressBarParseGames = new QProgressBar;
  progressBarInitializePatterns = new QProgressBar;
  progressBarInitializeGames->setMaximumHeight(20);
  progressBarDuplicateCheck->setMaximumHeight(20);
  progressBarParseGames->setMaximumHeight(20);
  progressBarInitializePatterns->setMaximumHeight(20);
  
  cancelButton = new QPushButton("&Cancel", this);
  backButton = new QPushButton("< &Back", this);
  finishButton = new QPushButton("&Finish", this);
  
  QGridLayout *layout = new QGridLayout;
  
  QGroupBox *groupBoxProgress = new QGroupBox(tr("Progress"));
  QGridLayout *gridLayoutProgress = new QGridLayout;
  gridLayoutProgress->addWidget(labelInitializeGames, 0, 0);
  gridLayoutProgress->addWidget(progressBarInitializeGames, 0, 1, 1, 2);
  gridLayoutProgress->addWidget(labelDuplicateCheck, 1, 0);
  gridLayoutProgress->addWidget(progressBarDuplicateCheck, 1, 1, 1, 2);
  gridLayoutProgress->addWidget(labelParseGames, 2, 0);
  gridLayoutProgress->addWidget(progressBarParseGames, 2, 1, 1, 2);
  gridLayoutProgress->addWidget(labelInitializePatterns, 3, 0);
  gridLayoutProgress->addWidget(progressBarInitializePatterns, 3, 1, 1, 2);
  groupBoxProgress->setLayout(gridLayoutProgress);
  groupBoxProgress->setMaximumHeight(150);
  groupBoxProgress->setFlat(true);
  layout->addWidget(groupBoxProgress, 0, 0, 1, 3);
  
  QGroupBox *groupBoxDetails = new QGroupBox(tr("Details"));
  QGridLayout *gridLayoutDetails = new QGridLayout;
  gridLayoutDetails->addWidget(labelCurrentTime, 0, 0, 1, 2);
  gridLayoutDetails->addWidget(labelCurrentTimeData, 0, 2);
  gridLayoutDetails->addWidget(labelTotalGames, 2, 0, 1, 2);
  gridLayoutDetails->addWidget(labelTotalGamesData, 2, 2);
  gridLayoutDetails->addWidget(labelDuplicateGames, 3, 0, 1, 2);
  gridLayoutDetails->addWidget(labelDuplicateGamesData, 3, 2);
  groupBoxDetails->setLayout(gridLayoutDetails);
  groupBoxDetails->setFlat(true);
  groupBoxDetails->setMaximumHeight(100);
  layout->addWidget(groupBoxDetails, 1, 0 , 1, 3);
  
  QSpacerItem* spacerButtons = new QSpacerItem(0, 0);
  QGridLayout *layoutButtons = new QGridLayout;
  QSpacerItem* spacer = new QSpacerItem(0, 0);
  layoutButtons->addItem(spacer, 0, 0);
  layoutButtons->addWidget(cancelButton, 0, 1);
  layoutButtons->addWidget(backButton, 0, 2);
  layoutButtons->addWidget(finishButton, 0, 3);
  layoutButtons->setColumnStretch(0, 1);
  layout->addItem(spacerButtons, 2, 0, 1, 3);
  layout->addItem(layoutButtons, 3, 0, 1, 3);  
  layout->setOriginCorner(Qt::TopLeftCorner);
  
  layout->setColumnMinimumWidth(0,100);
  layout->setColumnMinimumWidth(1,100);
  layout->setColumnMinimumWidth(2,100);
  setLayout(layout);

  connect(cancelButton, SIGNAL(clicked()),
          this, SLOT(cancelClick()));
  connect(backButton, SIGNAL(clicked()),
          this, SLOT(backClick()));
  connect(finishButton, SIGNAL(clicked()),
          addDatabase_, SLOT(accept()));

  finishButton->setEnabled(false);
  hasCancelled_ = false;
  isBackClicked_ = false;
}


void ProgressPage::setLabelTotalGamesData(u32 gamesTotal)
{
  QString data(lexical_cast<std::string>(gamesTotal).c_str());
  labelTotalGamesData->setText(data);
  repaint();
}



void ProgressPage::setLabelDuplicateGamesData(u32 duplicatesTotal)
{
  QString data(lexical_cast<std::string>(duplicatesTotal).c_str());
  labelDuplicateGamesData->setText(data);
  repaint();
}


// Back button is clicked 
// Flag to the Add database process that we need to exit and that back was selected
// If the Database was succesfully created then go straight to Back action
void ProgressPage::backClick()
{
  if (finishButton->isEnabled() == false) {
    hasCancelled_ = true;
    isBackClicked_ = true;
  }
  else if (finishButton->isEnabled() == true) {
    back();
  }
}


/***************************************************
  Go back to the SetupPage
******************************************************/
void ProgressPage::back()
{
//   std::cout << "back1" << std::endl;
  qApp->restoreOverrideCursor();
  progressBarInitializeGames->reset();
  progressBarDuplicateCheck->reset();
  progressBarParseGames->reset();
  progressBarInitializePatterns->reset();
  labelTotalGamesData->setText("");
  labelDuplicateGamesData->setText("");

  QString name = addDatabase_->setupPage->databaseNameLineEdit->text();
  std::string strName = name.toStdString();
  addDatabase_->mainWindow_->dbManager_->deleteDatabase(strName); 
  hasCancelled_ = false;
  isBackClicked_ = false;
  finishButton->setEnabled(false);
  addDatabase_->stackedWidget->setCurrentWidget(addDatabase_->setupPage);
  addDatabase_->repaint();
  addDatabase_->update();
}


// Cancel the Add Database Wizard and delete the database that was created
void ProgressPage::cancel()
{
//   qDebug("progresspage cancel()");
  qApp->restoreOverrideCursor();
  QString name = addDatabase_->setupPage->databaseNameLineEdit->text();
  std::string strName = name.toStdString();
  addDatabase_->mainWindow_->dbManager_->deleteDatabase(strName);
  addDatabase_->reject();
}



void ProgressPage::cancelClick()
{
  if (finishButton->isEnabled() == false) {
    hasCancelled_ = true; // To stop the DbManager from processing games 
  }
  else if (finishButton->isEnabled() == true) {
    cancel();
  }
}

