// Database.cpp 
// baduK++ 0.1 - Neil Mclean 
// Started 6/4/2006

#include <string>
#include <iostream>
#include <vector>
#include <map>
#include <fstream>
#include <cmath>
#include <QLabel>

#include "Database.h"
#include "LogFile.h"
#include "Timer.h"

#ifdef WIN32
  #include <C:\Boost\include\boost-1_33_1\boost\filesystem\operations.hpp>
  #include <C:\Boost\include\boost-1_33_1\boost\filesystem\fstream.hpp>
  #include <C:\Boost\include\boost-1_33_1\boost\filesystem\path.hpp>
#else
  #include "boost/filesystem/operations.hpp"
  #include "boost/filesystem/fstream.hpp"
#endif

typedef map<u32,boost::shared_ptr<Directory> > dirs;
typedef map<u32,boost::shared_ptr<Game> > GameIdx;
typedef multimap<u64,u32> FnvIdx; //FNV hash signature / Game Index
typedef map<u32,u32> GamesFormatTable;

using namespace std;
using boost::filesystem::path;
using boost::filesystem::native;
using boost::filesystem::directory_iterator;


Database::Database() {}

Database::Database(string name, string dirpath, DbManager* dbManager) : dbName_(name), 
                   dbDirpath_(dirpath), dbManager_(dbManager)
{
  path fullPath = system_complete( path(dirpath, native));
  if (exists(fullPath) == false) { 
    create_directory(fullPath); 
  }
  if (is_directory(fullPath) == false) { 
    create_directory(fullPath); 
  }
}



Database::~Database() 
{
  if (Games.empty() == false) { 
    Games.clear(); 
  }
  if (Dirs.empty() == false) { 
    Dirs.clear(); 
  }
  if (duplicates.empty() == false) { 
    duplicates.clear(); 
  }
  if (fnvSignatures.empty() == false) { 
    fnvSignatures.clear(); 
  }
}



//For finding the Dirpath from the Dir index associated with a game filename
string Database::getDir(const u32 index) const
{
  map<u32,boost::shared_ptr<Directory> >::const_iterator i;
  i = Dirs.find(index);
  return i->second->getDir();
}



//--------------------------------------------------
// Add Files & Directories function
//--------------------------------------------------
bool Database::addCollection(const string& dir, const bool isRecursive_) 
{
  if (addDirs(dir, isRecursive_) == true) { 
    map<u32,boost::shared_ptr<Directory> >::const_iterator i;
    for (i = Dirs.begin(); i != Dirs.end(); i++) {
      addFiles(i->first, i->second->getDir());    
    }
    gamesTotal_ = Games.size();
    return true;
  }
  return false;
}



// Check if the Directory is valid
bool Database::isDir(const string& dir) const
{
  path fullPath = system_complete(path(dir, native));
  if (exists(fullPath) == false) {
    cerr << "Error: the directory " << fullPath.string()
        << " does not exist.\n";
    return false;
  }
  if (is_directory(fullPath) == false) { // Is not a directory
    return false;
  }
  return true;
}



//-------------------------------------------------------------------
//! Add dirs to Database if they contain sgf files
//-----------------------------------------------------------------
bool Database::addDirs(const string& dir, const bool recursive)
{
  vector<string> dirpaths; // used for recursive dirs
  dirpaths.push_back(dir); //add passed argument directory to initialise vector
  bool hasSgf = false;
  do {
    path fullPath = complete(path(dirpaths.back(), native));
    dirpaths.pop_back();
    directory_iterator end;
    for (directory_iterator it(fullPath); it != end; ++it) { // Iterate dir 
      if ((is_directory(*it)) && (true == recursive)) {
        dirpaths.push_back(fullPath.string()+"/"+it->leaf());
      }
      else if (false == hasSgf) {
        size_t i = it->leaf().rfind('.', it->leaf().length());
        if ((i!=string::npos) &&
              (it->leaf().substr(i+1,it->leaf().length()-i) == "sgf")) { 
          hasSgf = true; 
        }
      }
    }
    if (true == hasSgf) { //Store Directory path as it contains sgf files
      hasSgf = false;
      char* s = strdup(fullPath.string().c_str());
      //dirsToProcess.push_back(s);
      try {
        boost::shared_ptr<Directory> DirPtr(new Directory(Dirs.size(), s));
        Dirs.insert(make_pair(Dirs.size(), DirPtr));
      }
      catch (bad_alloc&) { 
        cout << "Error allocating memory." << endl; 
        abort();
      }
    }
  }
  while (dirpaths.empty() == false);
  if (Dirs.empty() == true) { 
    return false; 
  }
  return true;
}



//Add files to the Games map
void Database::addFiles(const u32 idx, const string& dir) 
{ 
  path fullPath = complete(path(dir, native));
  directory_iterator end;
  for (directory_iterator it(fullPath); it != end; ++it) {  // Iterate dir
    if (!is_directory(*it)) { //file
      size_t i = it->leaf().rfind('.', it->leaf().length());
      if ((i!=string::npos) & 
        (it->leaf().substr(i+1,it->leaf().length()-i)=="sgf")) { 
        try {
          //! create new Game at last position, idx of dir, filename
          boost::shared_ptr<Game> IndexPtr(new Game(Games.size(), idx,it->leaf()));
          Games.insert (make_pair(Games.size(), IndexPtr));
        }
        catch (bad_alloc&) { 
          cout << "Error allocating memory." << endl; 
          abort();
        }
      }
    }
  } //while
}



u32 Database::getAllGamesTotal()
{
  return gamesTotal_;
}



string Database::getDatabaseDirPath()
{
  return dbDirpath_;
}



u32 Database::getGamesTotal()
{
  return gamesTotal_ - duplicatesTotal_;
}



u32 Database::getDuplicatesTotal()
{
  return duplicatesTotal_;
}


//----------------------------------------------------------------
//! Initialise Games indexes with sgf info 
//! get Game signature for duplicate check
//------------------------------------------------------
void Database::initializeGames()
{
  try {
    // GUI Progress timer in wizard
//     std::cout << "db init games" << std::endl;   
    dbManager_->mainWindow_->addDatabase_->progressPage->progressBarInitializeGames->setMinimum(0);
    dbManager_->mainWindow_->addDatabase_->progressPage->progressBarInitializeGames->setMaximum(Games.size());
//     std::cout << "db init games 2" << std::endl;   
  
    Timer timer(Games.size());
    //progress_display show_progress( Games.size() );
    GameIdx::const_iterator i;
    int j = 0;
  
    std::string dud;
    for (i = Games.begin(); i != Games.end(); i++) {
      i->second->initializeGame(getDir(i->second->getDirkey()));
      dbManager_->mainWindow_->addDatabase_->progressPage->progressBarInitializeGames->setValue(++j);
//       ++timer;
      if ((++timer % 100) == 0) { // every hundred records
      
//         timer.estimatedTime();
//         timer.currentTime();
//         timer.estimateTime();
        dbManager_->mainWindow_->addDatabase_->updateEvents();
//         dbManager_->mainWindow_->addDatabase_->updateEvents(INITIALIZE_GAMES, timer.dblEstimateTime());
        if (dbManager_->mainWindow_->addDatabase_->isCancelled() == true) {
          throw GuiCancelAddDatabase("User pressed cancel from Wizard");
        } 
        
      }
    }
  }
  catch (exception& e) { 
    cerr << e.what() << endl; 
    abort();
  }
}



//--------------------------------------------------------------
//Checks fnv signatures of games to ensure no matching sigantures
//creates a vector of duplicates ready to be erased from games map

// Need to pass back values of Games found, duplicates, zero move games
//---------------------------------------------------------------------
bool Database::duplicateCheck(bool eraseDuplicates)
{
  try {
//     LogFile logFile;
//     logFile.open();
  
    u32 zeroMoveGames = 0;
  
    u32 gamesTotal = 0;
    u32 duplicatesTotal = 0;
    dbManager_->mainWindow_->addDatabase_->progressPage->progressBarDuplicateCheck->setMinimum(0);
    dbManager_->mainWindow_->addDatabase_->progressPage->progressBarDuplicateCheck->setMaximum(Games.size());
    
    int j = 0;
    Timer timer(Games.size()); 
  
    GameIdx::const_iterator i;
    for (i = Games.begin(); i != Games.end(); i++) {
      //cout << "dup game=" << count << endl;
      if (i->second->getMoveCount()) {
        FnvIdx::const_iterator j = fnvSignatures.find(i->second->getFnvSignature());
        if (j == fnvSignatures.end()) { //not a duplicate
          fnvSignatures.insert(make_pair(i->second->getFnvSignature(),
                              i->second->getIndex()));
        }
        else {
  
          duplicates.push_back(i->first);
          duplicatesTotal++;
          GameIdx::const_iterator k = Games.find(j->second);
//           string id1 = (getDir(i->second->getDirkey()) + '/' + 
//               i->second->getFilename());
//           string id2 = (getDir(k->second->getDirkey()) + '/' + 
//               k->second->getFilename());
//           logFile.log("Omitted game : "+ id1);
//           logFile.log("Same as game : "+ id2);
        }
      }
      else {
        duplicates.push_back(i->first);
        zeroMoveGames++;
//         string id1 = (getDir(i->second->getDirkey())
//             + '/' + i->second->getFilename());
//         logFile.log("Game has no moves : "+ id1);
      }
      
      gamesTotal++; //count of games processed
      dbManager_->mainWindow_->addDatabase_->progressPage->progressBarDuplicateCheck->setValue(++j);
      
      if ((++timer % 100) == 0) { // every hundred records
//         dbManager_->mainWindow_->addDatabase_->updateEvents(DUPLICATE_CHECK,
//          timer.dblEstimateTime());
        dbManager_->mainWindow_->addDatabase_->updateEvents();
        if (dbManager_->mainWindow_->addDatabase_->isCancelled() == true) {
          throw GuiCancelAddDatabase("User pressed cancel from Wizard");
        }

      }
      
    } //for all games

    removeDuplicates(eraseDuplicates);
    
    // GUI display information 
    gamesTotal_ = gamesTotal;
    duplicatesTotal_ = duplicatesTotal;
  }
  catch (exception& e) { 
    cerr << e.what() << endl; 
    abort();
  }
  return true;
}



//---------------------------------------------------------
//remove duplicates found from games map
//! remove from media too? eg HDisk
//---------------------------------------------------------
void Database::removeDuplicates(bool eraseDuplicates) 
{
  bool shouldEraseDuplicates = false; 
  vector<u32>::const_iterator i;
  try {
    if (true == shouldEraseDuplicates) { //remove games from disk?
      for (i = duplicates.begin(); i != duplicates.end(); i++) {
        string id1 = (getDir(Games[*i]->getDirkey()) + '/' + 
                      Games[*i]->getFilename());
        path fullPath= system_complete(path(id1, native));
        remove(fullPath);
      }
    }
  }
  catch (exception& e) { 
    cerr << e.what() << endl; 
    abort();
  }
  //erase duplicate games
  if (true == eraseDuplicates) {
    for (i = duplicates.begin(); i != duplicates.end(); i++) {
      Games.erase(*i);
    }
  }
  duplicates.clear();
}



//---------------------------------------------------------------------
// convert all games into a faster replay format for pattern searches.
// also save a file with file offsets for each game into it.
//---------------------------------------------------------------------
void Database::parseGames()
{
// qDebug("in parse games");
  try {
    openOutFileStream( dbDirpath_+"/gamesForm.dat");
    if (gamesFormatIdx.empty() == false) { 
      gamesFormatIdx.clear();
    }
    dbManager_->mainWindow_->addDatabase_->progressPage->progressBarParseGames->setMinimum(0);
    dbManager_->mainWindow_->addDatabase_->progressPage->progressBarParseGames->setMaximum(Games.size());
    Timer timer(Games.size());
    int j = 0;
    u32 seekPos = 0; //For offset position into Index Table
    u32 origSeekPos;
    map<u32, boost::shared_ptr<Game> >::iterator i;
    for (i = Games.begin(); i != Games.end(); i++ ) { 
      // get WholeBoard patterns
      origSeekPos = seekPos;
      if (ofs_.is_open() == false) {
        qFatal("Cannot open fileStream for writing - Parse Games");
        throw ios_base::failure("cannot open file");
      }
      i->second->generateGameConversions( ofs_, seekPos );
	    gamesFormatIdx.insert( make_pair(i->second->getIndex(), origSeekPos ));
      
      dbManager_->mainWindow_->addDatabase_->progressPage->progressBarParseGames->setValue(++j);
      if ((++timer % 100) == 0) { // every hundred records
        dbManager_->mainWindow_->addDatabase_->updateEvents();
        if (dbManager_->mainWindow_->addDatabase_->isCancelled() == true) {
          throw GuiCancelAddDatabase("User pressed cancel from Wizard");
        }
      }
    }

// qDebug("in parse games 2");

    ofs_.close();
    openOutFileStream ( dbDirpath_ + "/gamesFormIdx.dat");
    writeGamesFormatIndex();
    ofs_.close();
  }
  catch (ios_base::failure& e) { 
	qDebug("i/o error catch");
    cout << "Error I/O failure " << e.what() << endl; 
    abort();
  }
  catch (exception& e) { 
    cerr << e.what( ) << endl; 
    abort();
  }
}



// Used to get index into Parsed games to get the moves of a game from the database
// Setup database Games Format Index
void Database::loadGamesFormatIndex()
{
  openInFileStream ( dbDirpath_ + "/gamesFormIdx.dat");
  readGamesFormatIndex();
  ifs_.close();
}





// Used by GUI to setup game moves to display on preview board
// Loads the moves for the game
// generates vector of Boards for each move to be used by the GUI to navigate between moves
void Database::initializeGame(u32 index) //Game Index
{
  try {
    openInFileStream(dbDirpath_ + "/gamesForm.dat");

    GameIdx::const_iterator i; //should point to my game index
    i = Games.find(index);
    GamesFormatTable::const_iterator j;

    j = gamesFormatIdx.find(index); // find the game record index
    if (j != gamesFormatIdx.end()) {
      ifs_.seekg(j->second); //goto it on file
      i->second->initializeGameMoves(ifs_);
      i->second->setupBoards();
    }
    ifs_.close();
  }
  catch (exception& e) {
    cerr << e.what() << endl; 
    abort();
  }
}



//-------------------------------------------------------
//! Generate Database patterns for all kinds of search
//-------------------------------------------------------
void Database::initializePatterns ()
{
  dbManager_->mainWindow_->addDatabase_->progressPage->progressBarInitializePatterns->setMinimum(0);
  dbManager_->mainWindow_->addDatabase_->progressPage->progressBarInitializePatterns->setMaximum(Games.size());
  Timer timer(Games.size());
  int k = 0;
  try{
    openInFileStream(dbDirpath_ + "/gamesForm.dat");
    GameIdx::const_iterator i;
    GamesFormatTable::const_iterator j;
    for (i = Games.begin(); i != Games.end(); i++) {
      j = gamesFormatIdx.find(i->first); // find the game record
      ifs_.seekg(j->second); //goto it on file
      i->second->initializeGameMoves(ifs_);
      WB.generatePattern(i->second);
      i->second->uninitializeGameMoves();

      dbManager_->mainWindow_->addDatabase_->progressPage->progressBarInitializePatterns->setValue(++k);
      if ((++timer % 100) == 0) { // every hundred records
        dbManager_->mainWindow_->addDatabase_->updateEvents();
        if (dbManager_->mainWindow_->addDatabase_->isCancelled() == true) {
          throw GuiCancelAddDatabase("User pressed cancel from Wizard");
        }
      }
    }
    ifs_.close();

    openOutFileStream(dbDirpath_ + "/WBentry.dat");
    WB.writeEntry(ofs_);
    ofs_.close();
    openOutFileStream(dbDirpath_ + "/WBpatterns.dat");
    WB.writePatterns(ofs_);
    ofs_.close();
    openOutFileStream(dbDirpath_ + "/WBgameslist.dat");
    WB.writeGamesList(ofs_);
    ofs_.close();
    
    // free resources used - memory
    WB.cleanupPatternGenerationResources();
  }
  catch (ios_base::failure& e) {
    cout << "Error I/O failure " << e.what() << endl; 
    abort();
  }
}



//-------------------------------------------------------
//! Search for a pattern in the Database
//  WholeBoard, HalfBoard, 9x9, 12x12, FreePattern, etc...
//  Only WholeBoard search at this time
//-------------------------------------------------------
vector<boost::shared_ptr<GamesList> > Database::search(const u16 dbIndex,
    const u16 moveNumber, const u64 value)
{
  vector<boost::shared_ptr<GamesList> > gameslist;
  try {
    openInFileStream(dbDirpath_+"/WBentry.dat");
    WB.readEntry(ifs_); // Read in table of Pattern Head entry points
    ifs_.close();
    openInFileStream(dbDirpath_+"/WBpatterns.dat");
    if (WB.searchPatterns(ifs_, moveNumber, value) == true) { //found
      ifs_.close(); // Uses 'seekpos' and 'hitcount' from searchPatterns
      openInFileStream(dbDirpath_+"/WBgameslist.dat");
      gameslist = WB.getGamesList(dbIndex, ifs_); 
    }
    ifs_.close();
  }
  catch (ios_base::failure& e) { 
    cout << "Error I/O failure " << e.what() << endl; 
    abort();
  }
  catch (exception& e) { 
    cerr << e.what( ) << endl; 
    abort();
  }
  return gameslist;
}




//----------------------------------------------------------------------
//! SECTION
//! Read & Write of Game Indexes, Directorys & Duplicate Signatures
//-------------------------------------------------------------------------
void Database::openDB() 
{
  try {
    openInFileStream(dbDirpath_ + "/fnvsigs.dat");
    if (fnvSignatures.empty() == false) { 
      fnvSignatures.clear(); 
    }
    readFnvSignatures();
    ifs_.close();
    openInFileStream(dbDirpath_ + "/games.dat");
    if (Games.empty() == true) {
      Games.clear(); 
    }
    readGame();
    ifs_.close();
    openInFileStream(dbDirpath_ + "/dirs.dat");
    if (Dirs.empty() == false) { 
      Dirs.clear(); 
    }
    readDirectory();
    ifs_.close();
  }
  catch (ios_base::failure& e) { 
    cout << "Error I/O failure " << e.what() << endl; 
    abort();
  }
  catch (exception& e) { 
    cerr << e.what( ) << endl; 
    abort();
  }
}



void Database::closeDB() 
{
  try {
    openOutFileStream(dbDirpath_ + "/fnvsigs.dat");
    writeFnvSignatures();
    ofs_.close();
    openOutFileStream(dbDirpath_ + "/games.dat");
    writeGame();
    ofs_.close();
    openOutFileStream(dbDirpath_ + "/dirs.dat");
    writeDirectory();
    ofs_.close();
  }
  catch (ios_base::failure& e) { 
    cout << "Error I/O failure " << e.what() << endl; 
    abort();
  }
  catch (exception& e) { 
    cerr << e.what() << endl; 
    abort();
  }
}



void Database::openInFileStream(const string& fileName) 
{
  ifs_.clear();
  ifs_.open(fileName.c_str(), ios_base::in | ios_base::binary);
  if (ifs_.is_open() == false) { 
    throw ios_base::failure("cannot open file" + fileName);
  }
}



void Database::openOutFileStream(const string& fileName) 
{
  ofs_.clear();
  ofs_.open(fileName.c_str(), ios_base::out | ios_base::binary | ios_base::trunc);
  if (ofs_.is_open() == false) {
    throw ios_base::failure("cannot open file" + fileName);
  }
}



//Write to file the < Game Index, File Offset > table
void Database::writeGamesFormatIndex() 
{
  u32 size = gamesFormatIdx.size();
  ofs_.write(reinterpret_cast<char *> (&size), sizeof size);
  GamesFormatTable::const_iterator i;
  for (i = gamesFormatIdx.begin(); i != gamesFormatIdx.end(); i++) {
    ofs_.write((char*) (&i->first), sizeof (i->first));
    ofs_.write((char*) (&i->second), sizeof (i->second));
  }
}



void Database::readGamesFormatIndex() 
{
  if (gamesFormatIdx.empty() == false) { 
    gamesFormatIdx.clear(); 
  }
  u32 size;
  ifs_.read(reinterpret_cast<char *> (&size), sizeof size);
  u32 idx;
  u32 seekPos;
  u32 i = 0;
  while ((i < size) && (!ifs_.eof())) {
    i++;
    ifs_.read(reinterpret_cast<char *> (&idx), sizeof (idx));
    ifs_.read(reinterpret_cast<char *> (&seekPos), sizeof (seekPos));
    gamesFormatIdx.insert(make_pair(idx, seekPos));
  }
}



//--------------------------------------
//! Fnv Hash Game Signatures DB Routines
//! should check how many to read in - eof is not a good way - best
//--------------------------------------
void Database::readFnvSignatures() 
{
  u32 size;
  ifs_.read(reinterpret_cast<char *> (&size), sizeof size);
  u32 sig;
  u32 idx;
  u32 i = 0;
  while ((i < size) && (!ifs_.eof())) {
    i++;
    ifs_.read(reinterpret_cast<char *> (&sig), sizeof sig);
    ifs_.read(reinterpret_cast<char *> (&idx), sizeof idx);
    fnvSignatures.insert(make_pair(sig, idx));
  }
}



void Database::writeFnvSignatures() 
{
  u32 size = fnvSignatures.size();
  ofs_.write(reinterpret_cast<char *> (&size), sizeof size);
  FnvIdx::iterator i = fnvSignatures.begin();
  for (; i != fnvSignatures.end(); i++) {
    ofs_.write((char*) (&i->first), sizeof (i->first));
    ofs_.write(reinterpret_cast<char *> (&i->second), sizeof (i->second));
  }
}



void Database::readGame() 
{ 
  u32 size;
  ifs_.read(reinterpret_cast<char *> (&size), sizeof size);
  u32 i = 0;
  while ((i < size) && (!ifs_.eof())) {
    i++;
    boost::shared_ptr<Game> game (new Game());
    game->readGameIndex(ifs_);
    Games.insert(make_pair(game->getIndex(), game));
  }
}



void Database::writeGame() 
{
  u32 size = Games.size();
  ofs_.write(reinterpret_cast<char *> (&size), sizeof size);
  GameIdx::const_iterator i = Games.begin();
  for (; i != Games.end(); i++) {
    i->second->writeGameIndex(ofs_);
  }
}



void Database::readDirectory() 
{
  u32 size;
  ifs_.read(reinterpret_cast<char *> (&size), sizeof size);
  u32 i = 0;
  while ((i < size) && (!ifs_.eof())) {
    i++;
    boost::shared_ptr<Directory> directory (new Directory());
    directory->readDir(ifs_);
    Dirs.insert(make_pair(directory->getIndex(),directory));
  }
}



void Database::writeDirectory() 
{
  u32 size = Dirs.size();
  ofs_.write(reinterpret_cast<char *> (&size), sizeof size);
  map<u32,boost::shared_ptr<Directory> >::const_iterator i;
  for (i = Dirs.begin(); i != Dirs.end(); i++) {
    i->second->writeDir(ofs_);
  }
}


